from os import path
from functools import partial

from docutils.core import publish_parts as rest
installed_markups = set(['rest'])
try:
    from markdown import markdown
    installed_markups.add('markdown')
except ImportError:
    pass
try:
    from textile import textile
    installed_markups.add('textile')
except ImportError:
    pass
try:
    import creole
    import creole2html
    installed_markups.add('creole')
except ImportError:
    pass

from swsg import NoninstalledPackage
from swsg.template_functions import clevercss

SUPPORTED_MARKUP_LANGUAGES = frozenset(
    ['rest', 'creole', 'textile', 'markdown'])


class UnsupportedMarkup(Exception):
    def __init__(self, markup):
        self.markup = markup

    def __str__(self):
        return 'the markup {0} is not supported'.format(self.markup)

    def __repr__(self):
        return '{0}({1})'.format(self.__class__.__name__, self.markup)


class BaseSource(object):
    def __init__(self, template_dir, default_template, text):
        splitted_text = text.split(u'\n', 2)
        try:
            first_line, second_line, rest = splitted_text
            first_lines = [first_line, second_line]
        except ValueError:
            # there is only one newline or no newline at all
            try:
                first_line, rest = splitted_text
                first_lines = [first_line]
            except ValueError:
                # there is no newline in the text, therefore there are no
                # variable assignments
                first_lines = []
                rest = text
        title = u''
        template = u''
        temp_first_lines = first_lines[:]
        for line in first_lines:
            # remove the line from the list of the first two lines where a
            # variable assignment can be found
            if line.startswith(u'title:'):
                title = line.lstrip(u'title:').strip()
                temp_first_lines.remove(line)
            elif line.startswith(u'template:'):
                template = line.lstrip(u'template:').strip()
                temp_first_lines.remove(line)
        # if the title is not set, "unknown" will be used. That means that
        # setting the title is highly recommended!
        self.title = title or u'unknown'
        # the directive "template" is optional. If it is set, its assigned
        # value is used as the template for this source. Otherwise, the default
        # template will be used
        self.template_path = path.join(
            template_dir,
            template or default_template)
        self.full_text = text
        self.text = u'\n'.join(temp_first_lines + [rest])

        try:
            rendered_source = self.render_templateless()
        except NotImplementedError:
            rendered_source = u''
        self.namespace = {
            'title': self.title,
            'content': rendered_source,
            'clevercss': clevercss}

    def __eq__(self, other):
        return type(self) == type(other) and self.full_text == other.full_text

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.full_text)

    def render_templateless(self):
        raise NotImplementedError

    def render(self, TemplateClass, **template_options):
        # render the template with the source's namespace
        with open(self.template_path) as fp:
            text = fp.read().decode('utf-8')
        template = TemplateClass(text)
        return template.render(self.namespace, **template_options)


class ReSTSource(BaseSource):
    def render_templateless(self):
        return rest(self.text, writer_name='html')['body']


class CreoleSource(BaseSource):
    def render_templateless(self):
        return creole2html.HtmlEmitter(creole.Parser(self.text).parse()).emit()


class TextileSource(BaseSource):
    def render_templateless(self):
        # the function textile.textile adds a tab character at the start of the
        # output, so we remove it to normalize the return value
        return textile(self.text).lstrip('\t')


class MarkdownSource(BaseSource):
    def render_templateless(self):
        return markdown(self.text, output_format='xhtml')


def get_source_class_by_markup(markup):
    markups = [
        (frozenset(('rest', 'rst')), ReSTSource),
        (frozenset(('creole',)), CreoleSource),
        (frozenset(('textile', 'tt')), TextileSource),
        (frozenset(('markdown', 'md')), MarkdownSource)]
    for markup_identifiers, SourceClass in markups:
        if markup in markup_identifiers:
            return SourceClass
    raise UnsupportedMarkup(markup)


def ensure_markup_is_valid_and_installed(markup_language):
    if markup_language not in SUPPORTED_MARKUP_LANGUAGES:
        raise ValueError(
            '{0} is not an allowed value for "markup_language"; '
            'possible valid values are: {1}'.format(
                markup_language, SUPPORTED_MARKUP_LANGUAGES))
    elif markup_language not in installed_markups:
        if markup_language == 'rest':
            missing_package = 'docutils'
        else:
            missing_package = markup_language
        raise NoninstalledPackage(missing_package)



from os import path

from clevercss import convert as convert_clevercss


def clevercssfile2cssfile(clevercssfile, cssfile):
    '''render the content of ``clevercssfile`` via ``clevercss.comvert`` and
    write its output into ``cssfile``'''
    ccss_text = clevercssfile.read()
    converted_ccss = convert_clevercss(ccss_text)
    cssfile.write(converted_ccss)


def clevercss(filename, css_filename=None, xhtml_style=False):
    if css_filename is None:
        css_filename = path.splitext(filename)[0] + '.css'
    with open(filename) as input:
        with open(css_filename, 'w') as output:
            clevercssfile2cssfile(input, output)
    return_code = '<link rel="stylesheet" type="text/css" href="{0}">'
    if xhtml_style:
        return_code = return_code[:-1] + ' />'
    return return_code.format(css_filename)



import string

SUPPORTED_TEMPLATE_ENGINES = frozenset(['simple', 'mako', 'jinja2', 'genshi'])

BASE_DEFAULT_TEMPLATE = '''<!DOCTYPE HTML>
<html>
  <head>
    <title>{title}</title>
    {head}
  </head>
  <body>
    {content}
  </body>
</html>'''

DEFAULT_SIMPLE_TEMPLATE = BASE_DEFAULT_TEMPLATE.format(
    title='${title}',
    head='<meta charset="utf-8">',
    content='${content}')

DEFAULT_MAKO_TEMPLATE = DEFAULT_SIMPLE_TEMPLATE

DEFAULT_GENSHI_TEMPLATE = BASE_DEFAULT_TEMPLATE.format(
    title='${title}',
    head='<meta charset="utf-8" />',
    content='${Markup(content)}')

DEFAULT_JINJA_TEMPLATE = BASE_DEFAULT_TEMPLATE.format(
    title='{{ title }}',
    head='<meta charset="utf-8">',
    content='{{ content }}')


class NonexistingSource(Exception):
    def __init__(self, source_path):
        self.source_path = source_path

    def __str__(self):
        return 'the source {0} does not exist'.format(self.source_path)

    def __repr__(self):
        return '{0}({1})'.format(self.__class__.__name__, self.source_path)


class UnsupportedTemplate(Exception):
    def __init__(self, template_language):
        self.template_language = template_language

    def __str__(self):
        return (
            'the template language {0} does either '
            'not exist or is not supported.').format(self.template_language)

    def __repr__(self):
        return '{0}({1})'.format(
            self.__class__.__name__, self.template_language)


class BaseTemplate(object):
    def __init__(self, text):
        '''
        Abstract base class for implementing template classes.
        '''
        self.text = text

    def __eq__(self, other):
        return (type(self) == type(other) and self.text == other.text)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.text) + hash(tuple(self.source_names))

    def render(self, namespace):
        raise NotImplementedError


class SimpleTemplate(BaseTemplate):
    'Render templates as described in :pep:`0292`'
    def render(self, namespace):
        template = string.Template(self.text)
        return template.safe_substitute(**namespace)


class MakoTemplate(BaseTemplate):
    def render(self, namespace):
        # import mako only here because this package is optional
        from mako.template import Template
        template = Template(self.text)
        return template.render(**namespace)


class Jinja2Template(BaseTemplate):
    def render(self, namespace, **options):
        # import jinja2 only here because this package is optional
        from jinja2 import Environment
        env = Environment(**options)
        template = env.from_string(self.text)
        return template.render(**namespace)


class GenshiTemplate(BaseTemplate):
    def render(self, namespace, **options):
        # import genshi only here because this package is optional
        from genshi.template.markup import MarkupTemplate
        template = MarkupTemplate(self.text)
        stream = template.generate(**namespace)
        # enforce conversion to unicode
        options['encoding'] = None
        rendered_template = stream.render(**options)
        return rendered_template


def get_template_class_by_template_language(template_language):
    normalized_template_language = template_language.lower()
    templates = [
        (frozenset(('simple',)), SimpleTemplate),
        (frozenset(('mako',)), MakoTemplate),
        (frozenset(('jinja', 'jinja2')), Jinja2Template),
        (frozenset(('genshi',)), GenshiTemplate),
    ]
    for template_identifiers, TemplateClass in templates:
        if normalized_template_language in template_identifiers:
            return TemplateClass
    raise UnsupportedTemplate(normalized_template_language)



import py.test
from swsg.templates import (BaseTemplate, SimpleTemplate,
    MakoTemplate, Jinja2Template, GenshiTemplate)

SOURCE_TEXT = u'title: source title\nsome **important** text'
SIMPLE_TEMPLATE_TEXT = u'<title>$title</title>$content'

FIRST_REAL_LINE = u'first real line of the template'


def test_template_equality():
    t1 = BaseTemplate(u'foo')
    t2 = BaseTemplate(u'foo')
    assert t1 == t2
    t3 = BaseTemplate(u'bar')
    assert t1 != t3
    assert t2 != t3


def test_simple_template():
    template = SimpleTemplate(SIMPLE_TEMPLATE_TEXT)
    title = 'test title'
    content = 'blah blah'
    namespace = {'title': title, 'content': content}
    rendered_template = template.render(namespace)
    assumed_result = SIMPLE_TEMPLATE_TEXT.replace(
        '$title', title).replace('$content', content)
    assert rendered_template == assumed_result


def test_mako_template():
    mako = py.test.importorskip('mako')
    # almost completely copied from mako's homepage
    template_text = u'''
<table>
    % for row in rows:
        ${makerow(row)}
    % endfor
</table>

<%def name="makerow(row)">
    <tr>
    % for name in row:
        <td>${name}</td>
    % endfor
    </tr>
</%def>'''.strip()
    rows = [list(range(10)) for row in range(10)]
    mako_template = MakoTemplate(template_text)
    rendered_mako_template = mako_template.render({'rows': rows})
    assumed_result = mako.template.Template(template_text).render(rows=rows)
    assert rendered_mako_template == assumed_result


def test_jinja2_template():
    jinja2 = py.test.importorskip('jinja2')
    template_text = u'''{% for item in r -%}
        {{ item }}
    {%- endfor %}'''
    jinja_template = Jinja2Template(template_text)
    rendered_jinja_template = jinja_template.render({'r': range(10)})
    assumed_result = jinja2.Template(template_text).render(r=list(range(10)))
    assert rendered_jinja_template == assumed_result


def test_genshi_template():
    py.test.importorskip('genshi')
    from genshi.template.markup import MarkupTemplate
    # the following snippet is copied from
    # http://genshi.edgewall.org/wiki/ApiDocs/genshi.template.markup
    # and extended by an assignment at the beginning
    #<ul py:with="items=range(10)">
    template_text = '''<div xmlns:py="http://genshi.edgewall.org/">
    <ul>
        <li py:for="item in items">${item}</li>
    </ul>
</div>'''
    genshi_template = GenshiTemplate(template_text)
    rendered_genshi_template = genshi_template.render({'items': range(10)})
    assumed_stream = MarkupTemplate(template_text).generate(items=range(10))
    assumed_result = assumed_stream.render()
    assert rendered_genshi_template == assumed_result



from functools import partial
from operator import is_
from hashlib import sha256

is_none = partial(is_, None)

def hash_file(filename):
    with open(filename) as fp:
        text = fp.read()
    return sha256(text).hexdigest()



#!/usr/bin/env python

from __future__ import print_function

import sys
import codecs
from os import makedirs, path, getcwd, name as operating_system
from itertools import imap, izip
from operator import itemgetter

from argparse import ArgumentParser
from texttable import Texttable
from py.io import TerminalWriter
from logbook import FileHandler, INFO, DEBUG

from swsg import __version__
from swsg.loggers import swsg_logger as logger
from swsg.file_paths import LOGFILE as DEFAULT_LOGFILE
from swsg.projects import (DEFAULT_SETTINGS, NonexistingProject, Project,
    list_project_instances, get_project_by_path, remove_project)
from swsg.sources import SUPPORTED_MARKUP_LANGUAGES
from swsg.templates import SUPPORTED_TEMPLATE_ENGINES
from swsg.utils import is_none


def get_logging_handler(args, logging_level):
    if args.logfile is None:
        if operating_system == 'nt':
            # FIXME: use the handler logbook.NTEventLogHandler for windows
            #        users
            raise OSError(
                'Windows is an operating system which is not supported yet. '
                'Install a POSIX compatible system and try again.')
        logfile = DEFAULT_LOGFILE
    else:
        logfile = args.logfile
    # make sure that the file exists
    path_, filename = path.split(logfile)
    if not path.exists(path_):
        makedirs(path_)
    if not path.exists(filename):
        with open(logfile, 'w'):
            pass
    return FileHandler(logfile, level=logging_level)


def format_list_of_projects():
    terminal_writer = TerminalWriter()
    terminal_width = terminal_writer.fullwidth
    table = Texttable(max_width=terminal_width)
    table.header(['Name', 'Path', 'Created', 'Last modified'])
    projects = list_project_instances()
    for p in projects:
        table.add_row([
            p.name, p.path,
            p.created.strftime('%c'), p.last_modified.strftime('%c')])
    return table.draw() if projects else "no project created yet"


def print_list_of_projects(args):
    print(format_list_of_projects())


def perform_quickstart(args):
    '''Ask the user for required information to initialize a project with the
    desired configuration automatically.
    '''
    def get_default_template_language():
        for option, default_value in DEFAULT_SETTINGS['general']:
            if option == 'template language':
                return default_value
    # the following instruction text is copied from
    #http://bitbucket.org/birkenfeld/sphinx/src/tip/sphinx/quickstart.py#cl-788
    print('''Please enter values for the following settings
(just press Enter to accept a default value, if one is given in brackets).''')
    prompts = [
        ('project directory', getcwd()),
        ('name of the project', None),
        ('template language', get_default_template_language())]
    answers = []
    for prompt, default_value in prompts:
        try:
            if default_value is None:
                full_prompt = '{0} (required!): '.format(prompt)
            else:
                full_prompt = '{0} [{1}]: '.format(prompt, default_value)
            answer = raw_input(full_prompt) or default_value
            # give the user a 2nd chance to enter a value if the default value
            # is None and the user hit enter anyway
            if answer is None:
                print('Error: the {0} is required'.format(prompt))
                answer = raw_input(full_prompt)
                if not answer:
                    raise EOFError
            answers.append(answer)
        except EOFError:
            sys.exit('interrupted')
    template_language = answers.pop()
    project = Project(*answers)
    template_specific_prompts = DEFAULT_SETTINGS.get(template_language, [])
    template_specific_answers = []
    for option, default_value in template_specific_prompts:
        try:
            prompt = '{0} [{1}]: '.format(option, default_value)
            answer = raw_input(prompt) or default_value
            template_specific_answers.append(answer)
        except EOFError:
            sys.exit('interrupted')
    project.init()
    config_items = izip(template_specific_prompts, template_specific_answers)
    defaultless_config_items = map(itemgetter(0), config_items)
    project.update_config(
        'general', [('template language', template_language)])
    project.update_config(template_language, defaultless_config_items)
    project.create_default_template()


def init_project(args):
    project = Project(args.project_directory, args.name)
    project.init()


def remove_project_(args):
    full_path = path.abspath(args.path)
    try:
        remove_project(full_path)
    except NonexistingProject, e:
        print(e, file=sys.stderr)
    else:
        path_, project_name = path.split(full_path)
        print('removed the project {0!r} from the directory {1!r}'.format(
            project_name, path_))


def validate_change_config(args):
    # at least one of the options must be given (otherwise, the file is only
    # touched but not edited)
    conf_values = [args.markup_language, args.template_language]
    if all(imap(is_none, conf_values)):
        print(
            'Error: Neither a markup language '
            'nor a template language was given.',
            file=sys.stderr)
        # the exit code 2 is used for CLI errors by convention
        sys.exit(2)


def change_config(args):
    # the project's directory is the current working directory
    project = get_project_by_path(getcwd())
    project.update_config(
        markup_language=args.markup_language,
        template_language=args.template_language)


def render(args):
    # the project's directory is the current working directory
    project = get_project_by_path(getcwd())
    for output_path, output in project.render():
        with codecs.open(output_path, 'w', 'utf-8') as fp:
            fp.write(output)


def parse_args(argv=sys.argv[1:]):
    parser = ArgumentParser()
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help=(
            'Enable verbosity mode (disabled per default). Log more logging '
            'messages as usual, but no debugging messages.'))
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help=(
            'Enable debugging mode (disabled per default). This option causes '
            'SWSG to log as many messages as possible and therefore overrides '
            'the option "verbose".'))
    parser.add_argument(
        '--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument('-l', '--logfile')
    subparsers = parser.add_subparsers()
    list_parser = subparsers.add_parser(
        'list-projects', help='List all projects in a fancy ASCII table.')
    list_parser.set_defaults(func=print_list_of_projects)
    quickstart_parser = subparsers.add_parser(
        'quickstart',
        help=('answer some questions to get a '
        'ready-to-use project with your desired settings'))
    quickstart_parser.set_defaults(func=perform_quickstart)
    init_parser = subparsers.add_parser(
        'init', help='create and initialize a new project')
    init_parser.add_argument('name', help='The name of the project')
    init_parser.add_argument(
        '-p', '--project-directory', default='.',
        help=(
            'The directory where the new project will be created. It must '
            'already exist before the calling this command.'))
    init_parser.set_defaults(func=init_project)
    remove_parser = subparsers.add_parser(
        'remove-project', help='remove a project which does already exist')
    remove_parser.add_argument(
        'path',
        help='The path to the project directory which will be removed')
    remove_parser.set_defaults(func=remove_project_)
    config_parser = subparsers.add_parser(
        'change-config',
        help=(
            'Change project-dependent configuration values. It is recommended '
            'to use this interface instead of editing the configuration file '
            'config.ini directly.'))
    config_parser.add_argument(
        '-m', '--markup-language', choices=SUPPORTED_MARKUP_LANGUAGES,
        help=(
            'The name of the markup language being used. Possible valid values'
            ' are: {0}'.format(', '.join(SUPPORTED_MARKUP_LANGUAGES))))
    config_parser.add_argument(
        '-t', '--template-language', choices=SUPPORTED_TEMPLATE_ENGINES,
        help=(
            'The name of the template engine being used. Possible valid values'
            ' are: {0}'.format(', '.join(SUPPORTED_TEMPLATE_ENGINES))))
    config_parser.set_defaults(func=change_config)
    render_parser = subparsers.add_parser(
        'render',
        help=(
            'Render the templates with their corresponding source files which '
            'are located in the current project directory.'))
    render_parser.set_defaults(func=render)
    return parser.parse_args(argv)


def set_logging_level(args, logger):
    if args.verbose:
        logger.level_name = INFO
    if args.debug:
        logger.level_name = DEBUG
    return logger


def main(argv=sys.argv[1:]):
    args = parse_args(argv)
    if args.func == change_config:
        validate_change_config(args)
    new_logger = set_logging_level(args, logger)
    handler = get_logging_handler(args, new_logger.level_name)
    with handler.applicationbound():
        args.func(args)

if __name__ == '__main__':
    main()



from functools import partial

import py.test
from swsg.sources import (UnsupportedMarkup, BaseSource, ReSTSource,
    CreoleSource, TextileSource, MarkdownSource, get_source_class_by_markup)


def test_base_source():
    Source = partial(BaseSource, '', 'default.html')
    source = Source('')
    assert source.title == 'unknown'
    assert source.template_path == 'default.html'
    assert source.text == ''
    source = Source('title: foo\nsome text')
    assert source.title == 'foo'
    assert source.template_path == 'default.html'
    assert source.text == 'some text'
    source = Source('some text')
    assert source.title == 'unknown'
    assert source.template_path == 'default.html'
    assert source.text == 'some text'
    source = Source('template: bar.html\nsome text')
    assert source.title == 'unknown'
    assert source.template_path == 'bar.html'
    assert source.text == 'some text'
    source = Source('title: foo\ntemplate: bar.html\nsome text')
    assert source.title == 'foo'
    assert source.template_path == 'bar.html'
    assert source.text == 'some text'
    # make sure that the order of assignments does not matter
    source = Source('template: bar.html\ntitle: foo\nsome text')
    assert source.title == 'foo'
    assert source.template_path == 'bar.html'
    assert source.text == 'some text'


def test_render_rest():
    r'''
>>> ReSTSource(u'simple test').render_templateless()
u'<p>simple test</p>\n'
>>> ReSTSource(u'**simple test**').render_templateless()
u'<p><strong>simple test</strong></p>\n'
>>> ReSTSource(u'`simple test`').render_templateless()
u'<p><cite>simple test</cite></p>\n'
>>> ReSTSource(u'``test``').render_templateless()
u'<p><tt class="docutils literal">test</tt></p>\n'
>>> ReSTSource(u'`example`_\n\n.. _example: http://example.com').render_templateless()
u'<p><a class="reference external" href="http://example.com">example</a></p>\n'
>>> ReSTSource(u'`example <http://example.com>`_').render_templateless()
u'<p><a class="reference external" href="http://example.com">example</a></p>\n'
    '''
    py.test.importorskip('docutils')


def test_render_creole():
    r'''
>>> CreoleSource(u'test').render_templateless()
u'<p>test</p>\n'
>>> CreoleSource(u'test\ntest').render_templateless()
u'<p>test test</p>\n'
    '''
    py.test.importorskip('creole')


def test_render_textile():
    r'''
>>> TextileSource('_This_ is a *test.*').render_templateless()
'<p><em>This</em> is a <strong>test.</strong></p>'
>>> TextileSource('* One\n* Two\n* Three').render_templateless()
'<ul>\n\t\t<li>One</li>\n\t\t<li>Two</li>\n\t\t<li>Three</li>\n\t</ul>'
>>> TextileSource('Link to "Slashdot":http://slashdot.org/').render_templateless()
'<p>Link to <a href="http://slashdot.org/">Slashdot</a></p>'
    '''
    py.test.importorskip('textile')


def test_render_markdown():
    r'''
>>> MarkdownSource('[Slashdot](http://slashdot.org/ "Slashdot - News for nerds, stuff that matters")').render_templateless()
u'<p><a href="http://slashdot.org/" title="Slashdot - News for nerds, stuff that matters">Slashdot</a></p>'
>>> MarkdownSource('---------------------------------------').render_templateless()
u'<hr />'
>>> MarkdownSource('![Python](http://python.org/community/logos/python-logo.png "The Python Logo")').render_templateless()
u'<p><img alt="Python" src="http://python.org/community/logos/python-logo.png" title="The Python Logo" /></p>'
    '''
    py.test.importorskip('markdown')


def test_get_source_class_by_markup():
    f = get_source_class_by_markup
    assert f('rest') == ReSTSource
    assert f('rst') == ReSTSource
    assert f('creole') == CreoleSource
    assert f('textile') == TextileSource
    assert f('tt') == TextileSource
    assert f('markdown') == MarkdownSource
    assert f('md') == MarkdownSource
    py.test.raises(UnsupportedMarkup, "f('does_not_exist')")
